C  C_SA: modified from original SIMMAN.f
C       For constrained problems: SA descent in X(1,nx), i.e. x-space 
C                                 SA ascent in X(nx,neq), i.e. lambda-space
C
C       Both are separated: (x,lambda) --> (x',lambda) 
C                           (x,lambda) --> (x,lambda')
C

      SUBROUTINE C_SA(NEQ,NC,rlambda,
     1              N,X,MAX,RT,EPS,NS,NT,NEPS,MAXEVL,LB,UB,C,IPRINT,
     2              ISEED1,ISEED2,T,VM,XOPT,FOPT,NACC,NFCNEV,NOBDS,IER,
     3              FSTAR,XP,NACP)

C  Type all external variables.
      DOUBLE PRECISION  X(*), LB(*), UB(*), C(*), VM(*), FSTAR(*),
     1                  XOPT(*), XP(*), T, EPS, RT, FOPT
      INTEGER  NACP(*), N, NS, NT, NEPS, NACC, MAXEVL, IPRINT,
     1         NOBDS, IER, NFCNEV, ISEED1, ISEED2
      LOGICAL  MAX

C  Type all internal variables.
      DOUBLE PRECISION  F, FP, P, PP, RATIO
      INTEGER  NUP, NDOWN, NREJ, NNEW, LNOBDS, H, I, J, M
      LOGICAL  QUIT

C  Type all functions.
      DOUBLE PRECISION  EXPREP
      REAL  RANMAR

C  Variables and functions for probabilistic lambda
      integer neq
      double precision  fmin, fval, vtol, rlambda(*)
      data vtol/1.0D-6/
      double precision  viomax, obj, cst 

C  Initialize the random number generator RANMAR.
      CALL RMARIN(ISEED1,ISEED2)

C  Set initial values.
      NACC = 0
      NOBDS = 0
      NFCNEV = 0
      IER = 99

      DO 10, I = 1, neq
         NACP(I) = 0
10    CONTINUE

      DO 20, I = 1, NEPS
         FSTAR(I) = 1.0D+20
20    CONTINUE 

C  If the initial temperature is not positive, notify the user and 
C  return to the calling routine.  
      IF (T .LE. 0.0) THEN
         WRITE(*,'(/,''  THE INITIAL TEMPERATURE IS NOT POSITIVE. ''
     1             /,''  RESET THE VARIABLE T. ''/)')
         IER = 3
         RETURN
      END IF

C  If the initial value is out of bounds, notify the user and return
C  to the calling routine.
      DO 30, I = 1, N
         IF ((X(I) .GT. UB(I)) .OR. (X(I) .LT. LB(I))) THEN
            CALL PRT1
            IER = 2
            RETURN
         END IF
30    CONTINUE

C  Evaluate the function with input X and return value as F.
      CALL FCN(N,X,F)

C  If the function is to be minimized, switch the sign of the function.
C  Note that all intermediate and final output switches the sign back
C  to eliminate any possible confusion for the user.
      IF(.NOT. MAX) F = -F
      NFCNEV = NFCNEV + 1
      FOPT = F
      FSTAR(1) = F
      IF(IPRINT .GE. 1) CALL PRT2(MAX,NEQ,X,F)

      if(viomax(x).LT.vtol) then
         fmin=obj(n,x)
         do 8 i=1,neq
            xopt(i)=x(i)
8        continue
      else
         fmin=1.0D+20
         do 9 i=1,neq
            xopt(i)=0.0D0
9        continue
      endif

C  Determine lambda's VIM by degree of violation
      do 6 i=1,nc
         rlambda(i)=1.0D0
6     continue
      do 7 i=n+1,neq
         vm(i)=rlambda(i-n)*cst((i-n),n,x)
7     continue

C  Start the main loop. Note that it terminates if (i) the algorithm
C  succesfully optimizes the function or (ii) there are too many
C  function evaluations (more than MAXEVL).
100   NUP = 0
      NREJ = 0
      NNEW = 0
      NDOWN = 0
      LNOBDS = 0

      DO 400, M = 1, NT
         DO 300, J = 1, NS
            DO 200, H = 1, neq

C  Generate XP, the trial value of X. Note use of VM to choose XP.
               DO 110, I = 1, neq
                  IF (I .EQ. H) THEN
                     if(i.GT.n) then
                        if(cst((i-n),n,x).LT.vtol) goto 200
                     endif
                     XP(I) = X(I) + (RANMAR()*2.- 1.) * VM(I)
                  ELSE
                     XP(I) = X(I)
                  END IF

C  If XP is out of bounds, select a point in bounds for the trial.
                  if(i.LE.n) then
                      IF((XP(I) .LT. LB(I)) .OR. (XP(I) .GT. UB(I))) THEN
                         XP(I) = LB(I) + (UB(I) - LB(I))*RANMAR()
                         LNOBDS = LNOBDS + 1
                         NOBDS = NOBDS + 1
                         IF(IPRINT .GE. 3) CALL PRT3(MAX,NEQ,XP,X,FP,F)
                      END IF
                  endif
110            CONTINUE

C  Evaluate the function with the trial point XP and return as FP.
               CALL FCN(N,XP,FP)
               IF(.NOT. MAX) FP = -FP
               NFCNEV = NFCNEV + 1
               IF(IPRINT .GE. 3) CALL PRT4(MAX,NEQ,XP,X,FP,F)

C  If too many function evaluations occur, terminate the algorithm.
               IF(NFCNEV .GE. MAXEVL) THEN
                  CALL PRT5
                  IF (.NOT. MAX) FOPT = -FOPT
                  IER = 1
                  RETURN
               END IF

C  Accept the new point if the function value increases.
               IF(h.LE.n.AND.fp.GE.f.OR.h.GT.n.AND.fp.LE.f) THEN
                  IF(IPRINT .GE. 3) THEN
                     WRITE(*,'(''  POINT ACCEPTED'')')
                  END IF
                  DO 120, I = 1, neq
                     X(I) = XP(I)
120               CONTINUE
                  F = FP
                  NACC = NACC + 1
                  NACP(H) = NACP(H) + 1
                  NUP = NUP + 1

C  If greater than any other point, record as new optimum.
                  if(h.LE.n.AND.viomax(xp).LT.vtol) then
                     fval=obj(n,xp)
                     if(fval.LT.fmin) then
                        IF(IPRINT .GE. 3) THEN
                           WRITE(*,'(''  NEW OPTIMUM'')')
                        END IF
                        DO 130, I = 1, neq
                           XOPT(I) = XP(I)
130                     CONTINUE
                        fmin=fval
                        fopt=fp
                        NNEW = NNEW + 1
                     end if
                  end if

C  If the point is lower, use the Metropolis criteria to decide on
C  acceptance or rejection.
               ELSE
                  if(h.LE.n) then
                     P = EXPREP((FP - F)/T)
                  else
                     P = EXPREP((F - FP)/T)
                  endif
                  PP = RANMAR()
                  IF(PP .LT. P) THEN
                     IF(IPRINT .GE. 3) CALL PRT6(MAX)
                     DO 140, I = 1, neq
                        X(I) = XP(I)
140                  CONTINUE
                     F = FP
                     NACC = NACC + 1
                     NACP(H) = NACP(H) + 1
                     NDOWN = NDOWN + 1
                  ELSE
                     NREJ = NREJ + 1
                     IF(IPRINT .GE. 3) CALL PRT7(MAX)
                  END IF
               END IF

200         CONTINUE
300      CONTINUE

C  Adjust VM so that approximately half of all evaluations are accepted.
         DO 310, I = 1, N
            RATIO = DFLOAT(NACP(I)) /DFLOAT(NS)
            IF (RATIO .GT. .6) THEN
               VM(I) = VM(I)*(1. + C(I)*(RATIO - .6)/.4)
            ELSE IF (RATIO .LT. .4) THEN
               VM(I) = VM(I)/(1. + C(I)*((.4 - RATIO)/.4))
            END IF
            IF (VM(I) .GT. (UB(I)-LB(I))) THEN
               VM(I) = UB(I) - LB(I)
            END IF
310      CONTINUE

         do 350 i=n+1,neq
            vm(i)=rlambda(i-n)*cst((i-n),n,x)
350      continue

         IF(IPRINT .GE. 2) THEN
            CALL PRT8(NEQ,VM,XOPT,X)
         END IF

         DO 320, I = 1, neq
            NACP(I) = 0
320      CONTINUE

400   CONTINUE

      IF(IPRINT .GE. 1) THEN
         CALL PRT9(MAX,NEQ,T,XOPT,VM,FOPT,NUP,NDOWN,NREJ,LNOBDS,NNEW)
      END IF

      do 55 i=1,n
         vm(i)=1.0D0
55    continue

      print *, " "
      print *, "Current T = ", T
      print *, "obj = ", obj(n,x),"    maximal violation=", viomax(x)
      print *, "x = ", (x(i),i=1,n)

C  Check termination criteria.
      QUIT = .FALSE.
      FSTAR(1) = F
      IF ((FOPT - FSTAR(1)) .LE. EPS) QUIT = .TRUE.
      DO 410, I = 1, NEPS
         IF (ABS(F - FSTAR(I)) .GT. EPS) QUIT = .FALSE.
410   CONTINUE
      if(t.LT.1.0D-12) quit=.TRUE.

C  Terminate SA if appropriate.
      IF (QUIT) THEN
         DO 420, I = 1, neq
            X(I) = XOPT(I)
420      CONTINUE
         IER = 0
         IF (.NOT. MAX) FOPT = -FOPT
         IF(IPRINT .GE. 1) CALL PRT10
         RETURN
      END IF

C  If termination criteria is not met, prepare for another loop.
      T = RT*T
      DO 430, I = NEPS, 2, -1
         FSTAR(I) = FSTAR(I-1)
430   CONTINUE
      if(nnew.GE.1) then
         F = FOPT
         DO 440, I = 1, neq
            X(I) = XOPT(I)
440      CONTINUE
      endif

C  Loop again.
      GO TO 100

      END

