C  population-based cSA 
C
C    After probablity descent in x for a number of steps, do
C    probability ascent in lambda

      SUBROUTINE pSA(rlambda,
     1              N,xx,MAX,RT,EPS,NS,NT,NEPS,MAXEVL,LB,UB,C,IPRINT,
     2              ISEED1,ISEED2,T,VM,XOPT,FOPT,NACC,NFCNEV,NOBDS,IER,
     3              FSTAR,xxp,NACP)

C  Parameter definition
      parameter(nx=2, nc=1)
      parameter(neq=nx+nc)

      parameter(nprt=3, nchd=3)
      parameter(ntot=nprt+nchd)

C  for population
      double precision xx(nprt,neq), xxp(ntot,neq)
      double precision ff(nprt), x(neq), xp(neq), y(neq)
      double precision avgvio(nc+1)

C  Type all external variables.
      DOUBLE PRECISION  LB(*), UB(*), C(*), VM(*), FSTAR(*),
     1                  XOPT(*), T, EPS, RT, FOPT
      INTEGER  NACP(*), N, NS, NT, NEPS, NACC, MAXEVL, IPRINT,
     1         NOBDS, IER, NFCNEV, ISEED1, ISEED2
      LOGICAL  MAX

C  Type all internal variables.
      DOUBLE PRECISION  F, FP, P, PP, RATIO
      INTEGER  NUP, NDOWN, NREJ, NNEW, LNOBDS, H, I, J, M
      LOGICAL  QUIT

C  Type all functions.
      DOUBLE PRECISION  EXPREP
      REAL  RANMAR

C  For cauchy distribution
      integer iscauchy
      double precision rateup, ratelow
      double precision fp0, fp1, xp0, xp1

C  For lambda
      integer ndecl(nc+1), nincl(nc+1)

C  For lambda's determinimistic rule
      integer neq
      double precision  fmin, fval, vtol, vl, rlambda(*)
      common /vtol/ vtol
      double precision  sh_cst, sh_viomax, sh_obj, xpi
      double precision  t0, xvio

      integer nreSA

      iscauchy=1
      nreSA=0
      t0=t

      if(iscauchy.EQ.0) then
         rateup=0.6D0
         ratelow=0.4D0
      else
         rateup=0.3D0
         ratelow=0.2D0
      endif

C  Initialize the random number generator RANMAR.
      CALL RMARIN(ISEED1,ISEED2)

C  Set initial values.
      NACC = 0
      NOBDS = 0
      NFCNEV = 0
      IER = 99

      DO 10, I = 1, N
         NACP(I) = 0
10    CONTINUE

      DO 20, I = 1, NEPS
         FSTAR(I) = 1.0D+20
20    CONTINUE 

C  If the initial temperature is not positive, notify the user and 
C  return to the calling routine.  
      IF (T .LE. 0.0) THEN
         WRITE(*,'(/,''  THE INITIAL TEMPERATURE IS NOT POSITIVE. ''
     1             /,''  RESET THE VARIABLE T. ''/)')
         IER = 3
         RETURN
      END IF

C  If the initial value is out of bounds, notify the user and return
C  to the calling routine.
      do 30 k=1, nprt
         do 35 j=1, nx
            IF ((xx(k,j) .GT. UB(j)) .OR. (xx(k,j) .LT. LB(j))) THEN
               CALL PRT1
               IER = 2
               RETURN
            END IF
35       continue
30    continue


C  Evaluate the function with input X and return value as F.
      FOPT = 0.0D0
      FSTAR(1) = 0.0D0 
      fmin=1.0D20
      do 25 j=1,neq
         xopt(j)=0.0D0
25    continue

      do 40 k=1,nprt
          do 45 j=1,neq
             x(j)=xx(k,j)
45        continue
          call fcn(nx,x,ff(k))
          if(sh_viomax(x).LT.vtol) then
             fval=sh_obj(n,x)
             if(fval.LT.fmin) then
                fmin=fval
                do 42 j=1,neq
                   xopt(j)=x(j)
42              continue
                fopt=ff(k)
             endif
          endif
40    continue

C  Determine lambda's VIM by degree of violation
      do 7 i=1,nc
         rlambda(i)=0.01D0
c         vm(i+n)=rlambda(i)*sh_cst(i,n,x)
7     continue

C  Start the main loop. Note that it terminates if (i) the algorithm
C  succesfully optimizes the function or (ii) there are too many
C  function evaluations (more than MAXEVL).
100   NUP = 0
      NREJ = 0
      NNEW = 0
      NDOWN = 0
      LNOBDS = 0

      do 110 i=1,nc
         ndecl(i)=0
         nincl(i)=0
110   continue

      DO 400, M = 1, NT

         call newpick(id,ff,xx,x,t)
         f=ff(id)

         do 350 i=1,neq
            xp(i)=x(i)
350      continue

         DO 300, J = 1, NS
            DO 200, H = 1, N

C  Generate XP, the trial value of X. Note use of VM to choose XP.
C  Evaluate the function with the trial point XP and return as FP.
               if(iscauchy.EQ.0) then
                  xp0=xpi(0,h,x(h),vm(h),lb(h),ub(h))
                  xp(h)=xp0
                  call fcn(n,xp,fp0)
                  fp=fp0
               else
                  xp1=xpi(1,h,x(h),vm(h),lb(h),ub(h))
                  xp(h)=xp1
                  call fcn(n,xp,fp1)
                  fp=fp1
               endif

               NFCNEV = NFCNEV + 1
               IF(IPRINT .GE. 3) CALL PRT4(MAX,NEQ,XP,X,FP,F)

C  Accept the new point if the function value increases.
               IF(FP .LT. F) THEN
                  IF(IPRINT .GE. 3) THEN
                     WRITE(*,'(''  POINT ACCEPTED'')')
                  END IF
                  X(h) = XP(h)
                  F = FP
                  NACC = NACC + 1
                  NACP(H) = NACP(H) + 1
                  NUP = NUP + 1

C  If greater than any other point, record as new optimum.
                  if(sh_viomax(x).LT.vtol) then
                     fval=sh_obj(n,x)
                     if(fval.LT.fmin) then
                        IF(IPRINT .GE. 3) THEN
                           WRITE(*,'(''  NEW OPTIMUM'')')
                        END IF
                        DO 130, I = 1, neq
                           XOPT(I) = X(I)
130                     CONTINUE
                        fmin=fval
                        fopt=fp
                        NNEW = NNEW + 1
                     end if
                  end if

C  If the point is lower, use the Metropolis criteria to decide on
C  acceptance or rejection.
               ELSE
                  P = EXPREP((F - FP)/T)
                  PP = RANMAR()
                  IF (PP .LT. P) THEN
                     IF(IPRINT .GE. 3) CALL PRT6(MAX)
                     X(h) = XP(h)
                     F = FP
                     NACC = NACC + 1
                     NACP(H) = NACP(H) + 1
                     NDOWN = NDOWN + 1
                  ELSE
                     xp(h)=x(h)
                     NREJ = NREJ + 1
                     IF(IPRINT .GE. 3) CALL PRT7(MAX)
                  END IF
               END IF

200         CONTINUE
300      CONTINUE

C  Adjust VM so that approximately half of all evaluations are accepted.
         DO 310, I = 1, N
            RATIO = DFLOAT(NACP(I)) /DFLOAT(NS)
            IF (RATIO .GT. rateup) THEN
               vm(I) = vm(I)*(1. + 3.5D0*C(I)*(RATIO - rateup)/(1.0D0-rateup))
            ELSE IF (RATIO .LT. ratelow) THEN
               vm(I) = vm(I)/(1. + C(I)*((ratelow - RATIO)/ratelow))
            END IF
            IF (vm(I) .GT. (UB(I)-LB(I))) vm(I) = UB(I) - LB(I)
310      CONTINUE

         do 80 i=1,nc
            avgvio(i)=0.0D0
            do 85 k=1,nprt
               do 90 j=1,nx
                  y(j)=xx(k,j)
90             continue
               vl=sh_cst(i,n,y)
               avgvio(i)=avgvio(i)+vl
85          continue
            avgvio(i)=avgvio(i)/nc
80       continue

         do 50 i=1,nc
            vl=avgvio(i)
            if(vl.LT.0.01D0*t) then
                ndecl(i)=ndecl(i)+1
                nincl(i)=0
                if(ndecl(i).GE.3) then
                    ndecl(i)=0
                    rlambda(i)=rlambda(i)*0.8D0
                    if(rlambda(i).LT.0.001D0) rlambda(i)=0.001D0
                endif
            else if(vl.GT.1.0D0*t) then
                nincl(i)=nincl(i)+1
                ndecl(i)=0
                if(nincl(i).GE.3) then
                    nincl(i)=0
                    rlambda(i)=rlambda(i)*1.25D0
                endif
            else
                ndecl(i)=0
                nincl(i)=0
            endif
c            vm(i+n)=rlambda(i)*vl
50       continue

         do 65 h=1,neq
            xp(h)=x(h)
65       continue

         do 60 h=n+1,neq
            vl=sh_cst((h-n),n,x)
            if(vl.LT.vtol) goto 60
            xp(h)=x(h)+(ranmar()*2.-1.)*rlambda(h-n)*vl
            if(xp(h).LT.0.0D0) xp(h)=0.0D0

            call fcn(n,xp,fp)
            if(fp.GT.f) then
               x(h)=xp(h)
               f=fp
            else
               p=exprep((fp-f)/T)
               pp=ranmar()
               if(pp.LT.p) then
                  x(h)=xp(h)
                  f=fp
               else
                  xp(h)=x(h)
               endif
            endif
60       continue

         ff(id)=f
         do 120 j=1,nx
             xx(id,j)=x(j)
120      continue
         do 125 k=1,nprt
            do 128 j=nx+1,neq
               xx(k,j)=x(j)
128         continue
125      continue

         DO 320, I = 1, N
            NACP(I) = 0
320      CONTINUE

400   CONTINUE

      do 55 i=1,n
         vm(i)=1.0D0
55    continue

c      print *, " "
c      print *, "Current T = ", T
c      print *, "obj = ", sh_obj(n,x),"    maximal violation=", sh_viomax(x)
c      print *, "x = ", (x(i),i=1,neq)

C  Check termination criteria.
      QUIT = .FALSE.
      FSTAR(1) = F
      IF ((FOPT - FSTAR(1)) .LE. EPS) QUIT = .TRUE.
      DO 410, I = 1, NEPS
         IF (ABS(F - FSTAR(I)) .GT. EPS) QUIT = .FALSE.
410   CONTINUE

      if(t.LT.1.0D-6.OR.quit) then
         xvio=sh_viomax(x)
         if(xvio.GT.vtol.AND.nreSA.GT.0) then
            nreSA=nreSA-1
            quit=.FALSE.
            t=xvio*100.0D0
            if(t.GT.t0) t=t0
            print *, "Re-annealing from T =", t
            goto 100
         else
            quit=.TRUE.
         endif
      endif

C  Terminate SA if appropriate.
      IF (QUIT) THEN
         IER = 0
         print *, "finish ppsa"
         call flush(6)
         RETURN
      END IF

C  If termination criteria is not met, prepare for another loop.
      T = RT*T
      DO 430, I = NEPS, 2, -1
         FSTAR(I) = FSTAR(I-1)
430   CONTINUE

c      if(nnew.GE.1) then
c         F = FOPT
c         DO 440, I = 1, N
c            X(I) = XOPT(I)
c440      CONTINUE
c      endif

C  Loop again.
      GO TO 100

      END

c
c  newpick: pick up a child from a population with the same lambda
c
      subroutine newpick(id,ff,xx,x,t)

      parameter(nx=2, nc=1, nprt=3)
      parameter(neq=nx+nc)

      integer id
      double precision ff(nprt), xx(nprt,neq), x(neq), t
      REAL RANMAR
      double precision EXPREP
      double precision p(nprt+1), psum, pp

      psum=0.0D0
      do 10 k=1, nprt
         p(k)=EXPREP(-(ff(k)-ff(1))/t)
         psum=psum+p(k)
10    continue
      do 20 k=1, nprt
         p(k)=p(k)/psum
20    continue

      pp=RANMAR()
      do 50 k=1,nprt
         if(pp.LE.p(k)) then
            id=k
            do 30 j=1,neq
               x(j)=xx(id,j)
30          continue
            return
         else
            p(k+1)=p(k)+p(k+1)
         endif
50    continue
      print *, "wrong in selection"
      print *, "p=", p, "  ff=", ff
      stop

      end

c
c  pickchild: pick up a child from a population
c
      subroutine pickchild(id,ff,xx,x,t)

      parameter(nx=2, nc=1, nprt=3)
      parameter(neq=nx+nc)

      integer id, kused(nprt), nrd
      double precision ff(nprt), xx(nprt,neq), t
      double precision x(neq), y01(neq), f0, f1, f01
      double precision pp, p, p0, p1 

      REAL RANMAR
      DOUBLE PRECISION EXPREP

      do 5 i=1,nprt
         kused(i)=0
5     continue

      k0=mod(irand(0),nprt)+1
      kused(k0)=1
      nrd=1

100   if(nrd.GE.nprt) then
          id=k0
          do 50 j=1,neq
             x(j)=xx(id,j)
50        continue
          return
      endif
20    k1=mod(irand(0),nprt)+1
      if(kused(k1).EQ.1) then
          goto 20
      else
          nrd=nrd+1
          kused(k1)=1
      endif

      do 25 j=1,neq
          if(j.LE.nx) then
              y01(j)=xx(k1,j)
          else
              y01(j)=xx(k0,j)
          endif
25    continue
      f0=ff(k0)
      f1=ff(k1)
      call fcn(nx,y01,f01)

      if(f01.LT.f0) then
          p0=1.0D0
      else
          p0=EXPREP((f0-f01)/t)
      endif
      if(f01.LT.f1) then
          p1=1.0D0
      else
          p1=EXPREP((f1-f01)/t)
      endif
      p=p0*p1
      pp = RANMAR()
      if(pp.LT.p) k0=k1
      goto 100
 
      end

