C  SA with stochastic Lambda rule
C
C    After probablity descent in x for a number of steps, do
C    probability ascent in lambda

      SUBROUTINE sa_equal(NEQ,NC,rlambda,xt,
     1              N,X,MAX,RT,EPS,NS,NT,NEPS,MAXEVL,LB,UB,C,IPRINT,
     2              ISEED1,ISEED2,T,VM,XOPT,FOPT,NACC,NFCNEV,NOBDS,IER,
     3              FSTAR,XP,NACP)

C  Type all external variables.
      DOUBLE PRECISION  X(*), LB(*), UB(*), C(*), VM(*), FSTAR(*),
     1                  XOPT(*), XP(*), T, EPS, RT, FOPT
      INTEGER  NACP(*), N, NS, NT, NEPS, NACC, MAXEVL, IPRINT,
     1         NOBDS, IER, NFCNEV, ISEED1, ISEED2
      LOGICAL  MAX

C  Type all internal variables.
      DOUBLE PRECISION  F, FP, P, PP, RATIO
      INTEGER  NUP, NDOWN, NREJ, NNEW, LNOBDS, H, I, J, M
      LOGICAL  QUIT

C  Type all functions.
      DOUBLE PRECISION  EXPREP
      REAL  RANMAR

C  For cauchy distribution
      integer iscauchy
      double precision rateup, ratelow

C  For false optimal points of relaxed constraints
      parameter(nx=2, ncc=1)
      parameter(neqq=nx+ncc)

      double precision ffopt, fxopt(neqq)
      integer ndecl(ncc+1), nincl(ncc+1)

C  For lambda's determinimistic rule
      integer neq
      double precision  fvtol, vtol
      data fvtol/1.0D-6/
      common /vtol/ vtol
      double precision  fmin, fval, vl, rlambda(*)
      double precision  sh_cst, sh_viomax, sh_obj, sh_oviomax, xpi
      double precision  xt(*), xvio, t0, fxt
      double precision  rleq, rtrleq
      common /rleq/ rleq

c  reannealing times
      integer nreSA
      logical redoNT

      iscauchy=1
      nreSA=10
      t0=t

c  ratio of reducing relaxed equality constraints
      rtrleq=0.95D0

      if(iscauchy.EQ.0) then
         rateup=0.6D0
         ratelow=0.4D0
      else
         rateup=0.3D0
         ratelow=0.2D0
      endif

C  Initialize the random number generator RANMAR.
      CALL RMARIN(ISEED1,ISEED2)

C  Set initial values.
      NACC = 0
      NOBDS = 0
      NFCNEV = 0
      IER = 99

      DO 10, I = 1, N
         NACP(I) = 0
10    CONTINUE

      DO 20, I = 1, NEPS
         FSTAR(I) = 1.0D+20
20    CONTINUE 

C  If the initial temperature is not positive, notify the user and 
C  return to the calling routine.  
      IF (T .LE. 0.0) THEN
         WRITE(*,'(/,''  THE INITIAL TEMPERATURE IS NOT POSITIVE. ''
     1             /,''  RESET THE VARIABLE T. ''/)')
         IER = 3
         RETURN
      END IF

C  If the initial value is out of bounds, notify the user and return
C  to the calling routine.
      DO 30, I = 1, N
         IF ((X(I) .GT. UB(I)) .OR. (X(I) .LT. LB(I))) THEN
            CALL PRT1
            IER = 2
            RETURN
         END IF
30    CONTINUE

C  Evaluate the function with input X and return value as F.
      CALL FCN(N,X,F)

C  If the function is to be minimized, switch the sign of the function.
C  Note that all intermediate and final output switches the sign back
C  to eliminate any possible confusion for the user.
      IF(.NOT. MAX) F = -F
      NFCNEV = NFCNEV + 1
      FOPT = F
      FSTAR(1) = F
      IF(IPRINT .GE. 1) CALL PRT2(MAX,NEQ,X,F)

      print *, "relaxed radius = ", rleq

      if(sh_viomax(x).LT.fvtol) then
         ffopt=f
         do 5 i=1,neq
            fxopt(i)=x(i)
5        continue
      endif

      if(sh_oviomax(x).LT.vtol) then
         fmin=sh_obj(n,x)
         do 8 i=1,neq
            xopt(i)=x(i)
8        continue
      else
         fmin=1.0D+20
         do 9 i=1,neq
            xopt(i)=0.0D0
9        continue
      endif

C  Determine lambda's VIM by degree of violation
      do 7 i=1,nc
         rlambda(i)=0.01D0
         vm(i+n)=rlambda(i)*sh_cst(i,n,x)
7     continue

      fxt=f
      do 70 i=1,neq
         xt(i)=x(i)
70    continue

C  Start the main loop. Note that it terminates if (i) the algorithm
C  succesfully optimizes the function or (ii) there are too many
C  function evaluations (more than MAXEVL).

100   NUP = 0
      NREJ = 0
      NNEW = 0
      NDOWN = 0
      LNOBDS = 0

      do 110 i=1,nc
         ndecl(i)=0
         nincl(i)=0
110   continue

      DO 400, M = 1, NT

         redoNT=.FALSE.

         do 350 i=1,neq
            xp(i)=x(i)
350      continue

         DO 300, J = 1, NS
            DO 200, H = 1, N

C  Generate XP, the trial value of X. Note use of VM to choose XP.
C  Evaluate the function with the trial point XP and return as FP.

               xp(h)=xpi(iscauchy,h,x(h),vm(h),lb(h),ub(h))
               call fcn(n,xp,fp)

               IF(.NOT. MAX) FP = -FP
               NFCNEV = NFCNEV + 1
               IF(IPRINT .GE. 3) CALL PRT4(MAX,NEQ,XP,X,FP,F)

C  If too many function evaluations occur, terminate the algorithm.
c               IF(NFCNEV .GE. MAXEVL) THEN
c                  CALL PRT5
c                  IF (.NOT. MAX) FOPT = -FOPT
c                  IER = 1
c                  RETURN
c               END IF

C  Accept the new point if the function value increases.
               IF(FP .GE. F) THEN
                  IF(IPRINT .GE. 3) THEN
                     WRITE(*,'(''  POINT ACCEPTED'')')
                  END IF
                  X(h) = XP(h)
                  F = FP
                  NACC = NACC + 1
                  NACP(H) = NACP(H) + 1
                  NUP = NUP + 1

C  If greater than any other point, record as new optimum.
                  if(sh_viomax(x).LT.fvtol.AND.rleq.GT.1.0D-6) then
                     redoNT=.TRUE.
                     rleq=rleq*rtrleq
                     fxt=f
                     ffopt=f
                     do 71 i=1,neq
                         xt(i)=x(i)
                         fxopt(i)=x(i)
71                   continue
                     nnew=nnew+1
                     print *, "relaxed radius reduced to rleq = ", rleq
                     call flush(6)
                  endif
                  if(sh_oviomax(x).LT.vtol) then
                     fval=sh_obj(n,x)
                     if(fval.LT.fmin) then
                        IF(IPRINT .GE. 3) THEN
                           WRITE(*,'(''  NEW OPTIMUM'')')
                        END IF
                        DO 130, I = 1, neq
                           fxopt(i) = x(i)
                           XOPT(I) = X(I)
130                     CONTINUE
                        fmin=fval
                        ffopt=fp
                        fopt=fp
                        NNEW = NNEW + 1
                     end if
                  end if

C  If the point is lower, use the Metropolis criteria to decide on
C  acceptance or rejection.
               ELSE
                  P = EXPREP((FP - F)/T)
                  PP = RANMAR()
                  IF (PP .LT. P) THEN
                     IF(IPRINT .GE. 3) CALL PRT6(MAX)
                     X(h) = XP(h)
                     F = FP
                     NACC = NACC + 1
                     NACP(H) = NACP(H) + 1
                     NDOWN = NDOWN + 1
                  ELSE
                     xp(h)=x(h)
                     NREJ = NREJ + 1
                     IF(IPRINT .GE. 3) CALL PRT7(MAX)
                  END IF
               END IF

200         CONTINUE
300      CONTINUE

C  Adjust VM so that approximately half of all evaluations are accepted.
         DO 310, I = 1, N
            RATIO = DFLOAT(NACP(I)) /DFLOAT(NS)
            IF (RATIO .GT. rateup) THEN
               vm(I) = vm(I)*(1. + 3.5D0*C(I)*(RATIO - rateup)/(1.0D0-rateup))
            ELSE IF (RATIO .LT. ratelow) THEN
               vm(I) = vm(I)/(1. + C(I)*((ratelow - RATIO)/ratelow))
            END IF
            IF (vm(I) .GT. (UB(I)-LB(I))) vm(I) = UB(I) - LB(I)
310      CONTINUE

         do 50 i=1,nc
            vl=sh_cst(i,n,x)
            if(vl.LT.0.01D0*t) then
                ndecl(i)=ndecl(i)+1
                nincl(i)=0
                if(ndecl(i).GE.3) then
                    ndecl(i)=0
                    rlambda(i)=rlambda(i)*0.8D0
                    if(rlambda(i).LT.0.001D0) rlambda(i)=0.001D0
                endif
            else if(vl.GT.1.0D0*t) then
                nincl(i)=nincl(i)+1
                ndecl(i)=0
                if(nincl(i).GE.3) then
                    nincl(i)=0
                    rlambda(i)=rlambda(i)*1.25D0
                endif
            else
                ndecl(i)=0
                nincl(i)=0
            endif
            vm(i+n)=rlambda(i)*vl
50       continue

         do 65 h=1,neq
            xp(h)=x(h)
65       continue

         do 60 h=n+1,neq
            if(sh_cst((h-n),n,x).LT.vtol) goto 60
            xp(h)=x(h)+(ranmar()*2.-1.)*vm(h)
            if(xp(h).LT.0.0D0) xp(h)=0.0D0

            call fcn(n,xp,fp)
            if(.NOT.max) fp=(-fp)
            if(fp.LE.f) then
               x(h)=xp(h)
               f=fp
            else
               p=exprep((f-fp)/T)
               pp=ranmar()
               if(pp.LT.p) then
                  x(h)=xp(h)
                  f=fp
               else
                  xp(h)=x(h)
               endif
            endif
60       continue

         IF(IPRINT .GE. 2) THEN
            CALL PRT8(NEQ,VM,XOPT,X)
         END IF

         DO 320, I = 1, N
            NACP(I) = 0
320      CONTINUE

         if(redoNT) then
            print *, "redo NT"
            goto 100
         endif

400   CONTINUE

      IF(IPRINT .GE. 1) THEN
         CALL PRT9(MAX,NEQ,T,XOPT,VM,FOPT,NUP,NDOWN,NREJ,LNOBDS,NNEW)
      END IF

      do 55 i=1,n
         vm(i)=Min(1.0D0,(ub(i)-lb(i)))
55    continue

C  Check termination criteria.
      QUIT = .FALSE.
      FSTAR(1) = F
      IF ((FOPT - FSTAR(1)) .LE. EPS) QUIT = .TRUE.
      DO 410, I = 1, NEPS
         IF (ABS(F - FSTAR(I)) .GT. EPS) QUIT = .FALSE.
410   CONTINUE

      if(t.LT.1.0D-6.OR.quit) then
          xvio=sh_oviomax(x)
          if(xvio.GT.ovtol.AND.nreSA.GT.0) then
              nreSA=nreSA-1
              quit=.FALSE.
              t=rleq*100.0D0
              if(t.GT.t0) t=t0
              f=fxt
              do 72 i=1,neq
                  x(i)=xt(i)
72            continue
              print *, "Re-annealing from T =", t, " and rleq =", rleq
              goto 100
          else
              quit=.TRUE.
          endif
      endif

C  Terminate SA if appropriate.
      IF (QUIT) THEN
C         DO 420, I = 1, N
C            Xopt(I) = X(I)
C420      CONTINUE
         IER = 0
         IF (.NOT. MAX) FOPT = -FOPT
         IF(IPRINT .GE. 1) CALL PRT10
         RETURN
      END IF

C  If termination criteria is not met, prepare for another loop.
      T = RT*T
      DO 430, I = NEPS, 2, -1
         FSTAR(I) = FSTAR(I-1)
430   CONTINUE

      if(nnew.GE.1) then
         f = ffopt
         DO 440, I = 1, N
            X(I) = fxopt(I)
440      CONTINUE
      endif

C  Loop again.
      GO TO 100

      END

