C  SA with stochastic Lambda rule
C
C    After probablity descent in x for a number of steps, do
C    probability ascent in lambda

      SUBROUTINE sa_local(NEQ,NC,rlambda,xdon,
     1              N,X,MAX,RT,EPS,NS,NT,NEPS,MAXEVL,LB,UB,C,IPRINT,
     2              ISEED1,ISEED2,T,VM,XOPT,FOPT,NACC,NFCNEV,NOBDS,IER,
     3              FSTAR,XP,NACP)

C  Type all external variables.
      DOUBLE PRECISION  X(*), LB(*), UB(*), C(*), VM(*), FSTAR(*),
     1                  XOPT(*), XP(*), T, EPS, RT, FOPT
      INTEGER  NACP(*), N, NS, NT, NEPS, NACC, MAXEVL, IPRINT,
     1         NOBDS, IER, NFCNEV, ISEED1, ISEED2
      LOGICAL  MAX

C  Type all internal variables.
      DOUBLE PRECISION  F, FP, P, PP, RATIO
      INTEGER  NUP, NDOWN, NREJ, NNEW, LNOBDS, H, I, J, M
      LOGICAL  QUIT

C  Type all functions.
      DOUBLE PRECISION  EXPREP
      REAL  RANMAR

C  For lambda's determinimistic rule
      integer neq
      double precision  fmin, fval, vtol, vl, rlambda(*)
      data vtol/1.0D-6/
      double precision  cst, viomax, obj
      double precision  xdon(*), xvio, t0

      t0=t

C  Initialize the random number generator RANMAR.
      CALL RMARIN(ISEED1,ISEED2)

C  Set initial values.
      NACC = 0
      NOBDS = 0
      NFCNEV = 0
      IER = 99

      DO 10, I = 1, N
         NACP(I) = 0
10    CONTINUE

      DO 20, I = 1, NEPS
         FSTAR(I) = 1.0D+20
20    CONTINUE 

C  If the initial temperature is not positive, notify the user and 
C  return to the calling routine.  
      IF (T .LE. 0.0) THEN
         WRITE(*,'(/,''  THE INITIAL TEMPERATURE IS NOT POSITIVE. ''
     1             /,''  RESET THE VARIABLE T. ''/)')
         IER = 3
         RETURN
      END IF

C  If the initial value is out of bounds, notify the user and return
C  to the calling routine.
      DO 30, I = 1, N
         IF ((X(I) .GT. UB(I)) .OR. (X(I) .LT. LB(I))) THEN
            CALL PRT1
            IER = 2
            RETURN
         END IF
30    CONTINUE

C  Evaluate the function with input X and return value as F.
      CALL FCN(N,X,F)

C  If the function is to be minimized, switch the sign of the function.
C  Note that all intermediate and final output switches the sign back
C  to eliminate any possible confusion for the user.
      IF(.NOT. MAX) F = -F
      NFCNEV = NFCNEV + 1
      FOPT = F
      FSTAR(1) = F
      IF(IPRINT .GE. 1) CALL PRT2(MAX,NEQ,X,F)

      if(viomax(x).LT.vtol) then
         fmin=obj(n,x)
         do 8 i=1,neq
            xopt(i)=x(i)
8        continue
      else
         fmin=1.0D+20
         do 9 i=1,neq
            xopt(i)=0.0D0
9        continue
      endif

C  Determine lambda's VIM by degree of violation
      do 7 i=1,nc
         rlambda(i)=0.01D0
         vm(i+n)=rlambda(i)*cst(i,n,x)
7     continue

      nbig_vio=0

C  Start the main loop. Note that it terminates if (i) the algorithm
C  succesfully optimizes the function or (ii) there are too many
C  function evaluations (more than MAXEVL).
100   NUP = 0
      NREJ = 0
      NNEW = 0
      NDOWN = 0
      LNOBDS = 0

      ndecl=0
      nincl=0

      DO 400, M = 1, NT

         do 350 i=n+1,neq
            xp(i)=x(i)
350      continue

         DO 300, J = 1, NS
            DO 200, H = 1, N

C  Generate XP, the trial value of X. Note use of VM to choose XP.
               DO 110, I = 1, N
                  IF (I .EQ. H) THEN
                     XP(I) = X(I) + (RANMAR()*2.- 1.) * VM(I)
                  ELSE
                     XP(I) = X(I)
                  END IF

C  If XP is out of bounds, select a point in bounds for the trial.
                  IF((XP(I) .LT. LB(I)) .OR. (XP(I) .GT. UB(I))) THEN
                    XP(I) = LB(I) + (UB(I) - LB(I))*RANMAR()
                    LNOBDS = LNOBDS + 1
                    NOBDS = NOBDS + 1
                    IF(IPRINT .GE. 3) CALL PRT3(MAX,NEQ,XP,X,FP,F)
                  END IF
110            CONTINUE

C  Evaluate the function with the trial point XP and return as FP.
               CALL FCN(N,XP,FP)
               IF(.NOT. MAX) FP = -FP
               NFCNEV = NFCNEV + 1
               IF(IPRINT .GE. 3) CALL PRT4(MAX,NEQ,XP,X,FP,F)

C  If too many function evaluations occur, terminate the algorithm.
               IF(NFCNEV .GE. MAXEVL) THEN
                  CALL PRT5
                  IF (.NOT. MAX) FOPT = -FOPT
                  IER = 1
                  RETURN
               END IF

C  Accept the new point if the function value increases.
               IF(FP .GE. F) THEN
                  IF(IPRINT .GE. 3) THEN
                     WRITE(*,'(''  POINT ACCEPTED'')')
                  END IF
                  DO 120, I = 1, N
                     X(I) = XP(I)
120               CONTINUE
                  F = FP
                  NACC = NACC + 1
                  NACP(H) = NACP(H) + 1
                  NUP = NUP + 1

C  If greater than any other point, record as new optimum.
                  if(viomax(x).LT.vtol) then
                     fval=obj(n,x)
                     if(fval.LT.fmin) then
                        IF(IPRINT .GE. 3) THEN
                           WRITE(*,'(''  NEW OPTIMUM'')')
                        END IF
                        DO 130, I = 1, neq
                           XOPT(I) = X(I)
130                     CONTINUE
                        fmin=fval
                        fopt=fp
                        NNEW = NNEW + 1
                     end if
                  end if

C  If the point is lower, use the Metropolis criteria to decide on
C  acceptance or rejection.
               ELSE
                  P = EXPREP((FP - F)/T)
                  PP = RANMAR()
                  IF (PP .LT. P) THEN
                     IF(IPRINT .GE. 3) CALL PRT6(MAX)
                     DO 140, I = 1, N
                        X(I) = XP(I)
140                  CONTINUE
                     F = FP
                     NACC = NACC + 1
                     NACP(H) = NACP(H) + 1
                     NDOWN = NDOWN + 1
                  ELSE
                     NREJ = NREJ + 1
                     IF(IPRINT .GE. 3) CALL PRT7(MAX)
                  END IF
               END IF

200         CONTINUE
300      CONTINUE

C  Adjust VM so that approximately half of all evaluations are accepted.
         DO 310, I = 1, N
            RATIO = DFLOAT(NACP(I)) /DFLOAT(NS)
            IF (RATIO .GT. .6) THEN
               VM(I) = VM(I)*(1. + C(I)*(RATIO - .6)/.4)
            ELSE IF (RATIO .LT. .4) THEN
               VM(I) = VM(I)/(1. + C(I)*((.4 - RATIO)/.4))
            END IF
            IF (VM(I) .GT. (UB(I)-LB(I))) THEN
               VM(I) = UB(I) - LB(I)
            END IF
310      CONTINUE

         if(t.LT.1.0D0.AND.t.LT.(1.0D-4*t0)) then
            call gdesc_inx(2,x)
            call fcn(n,x,f)
            if(.NOT.max) f=(-f)
         endif

         do 50 i=1,nc
            vl=cst(i,n,x)
            if(t.GT.1.0D-3) then
                if(vl.LT.0.001D0*t) then
                    ndecl=ndecl+1
                    nincl=0
                    if(ndecl.GE.3) then
                        ndecl=0
                        rlambda(i)=rlambda(i)*0.8D0
                    endif
                else if(vl.GT.0.1D0*t) then
                    nincl=nincl+1
                    ndecl=0
                    if(nincl.GE.2) then
                        nincl=0
                        rlambda(i)=rlambda(i)*2.0D0
                    endif
                else
                    ndecl=0
                    nincl=0
                endif
            else
                rlambda(i)=1.0D0
            endif
c            rlambda(i)=0.1D0
            vm(i+n)=rlambda(i)*vl
50       continue

         do 65 h=1,n
            xp(h)=x(h)
65       continue

         do 60 h=n+1,neq
            if(cst((h-n),n,x).LT.vtol) goto 60
            do 61 i=n+1,neq
               if(i.EQ.h) then
                  xp(i)=x(i)+(ranmar()*2.-1.)*vm(i)
c                  if(xp(i).LT.0.0D0) xp(i)=0.0D0
               else
                  xp(i)=x(i)
               endif
61          continue

            call fcn(n,xp,fp)
            if(.NOT.max) fp=(-fp)
            if(fp.LE.f) then
               do 62 i=n+1,neq
                   x(i)=xp(i)
62             continue
               f=fp
            else
               p=exprep((f-fp)/T)
               pp=ranmar()
               if(pp.LT.p) then
                  do 63 i=n+1,neq
                     x(i)=xp(i)
63                continue
                  f=fp
               endif
            endif
60       continue

         IF(IPRINT .GE. 2) THEN
            CALL PRT8(NEQ,VM,XOPT,X)
         END IF

         DO 320, I = 1, N
            NACP(I) = 0
320      CONTINUE

400   CONTINUE

      IF(IPRINT .GE. 1) THEN
         CALL PRT9(MAX,NEQ,T,XOPT,VM,FOPT,NUP,NDOWN,NREJ,LNOBDS,NNEW)
      END IF

      do 55 i=1,n
         vm(i)=1.0D0
55    continue

      print *, " "
      print *, "Current T = ", T
      xvio=viomax(x)
      print *, "obj = ", obj(n,x),"    maximal violation=", xvio
      print *, "x = ", (x(i),i=1,n)

C  Check termination criteria.
      QUIT = .FALSE.
      FSTAR(1) = F
      IF ((FOPT - FSTAR(1)) .LE. EPS) QUIT = .TRUE.
      DO 410, I = 1, NEPS
         IF (ABS(F - FSTAR(I)) .GT. EPS) QUIT = .FALSE.
410   CONTINUE
      if(t.LT.1.0D-5) quit=.TRUE.

C  Terminate SA if appropriate.
      IF (QUIT) THEN
c         DO 420, I = 1, N
c            X(I) = XOPT(I)
c420      CONTINUE
         IER = 0
         IF (.NOT. MAX) FOPT = -FOPT
         IF(IPRINT .GE. 1) CALL PRT10
         RETURN
      END IF

C  If termination criteria is not met, prepare for another loop.
      T = RT*T
      DO 430, I = NEPS, 2, -1
         FSTAR(I) = FSTAR(I-1)
430   CONTINUE

      if(xvio.GT.1.0D-3.AND.xvio.GT.10.0D0*t) then
         nbig_vio=nbig_vio+1
      else
         nbig_vio=0
      endif

      if(nbig_vio.GE.10) then
         do 70 i=1,neq
            xdon(i)=x(i)
70       continue

         call by_donlp2(msol,xdon,xobj,xvio,ni_used)

         if(msol.EQ.1.AND.xobj.LT.fmin) then
            fmin=xobj
            call fcn(n,xdon,fp)
            if(.NOT.max) fp=(-fp)
            fopt=fp
            do 75 i=1,neq
               xopt(i)=xdon(i)
75          continue
            nbig_vio=0
            nnew=nnew+1
         endif
      endif

      if(nnew.GE.1) then
         F = FOPT
         DO 440, I = 1, N
            X(I) = XOPT(I)
440      CONTINUE
      endif

C  Loop again.
      GO TO 100

      END

